#!/usr/bin/perl -w
#
#   Copyright 2012 Andrii Golovynskyi
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

#
# Module for interacting with ARC 1+ middleware

use strict;

use DBI;
use DBD::mysql;

use Expect;
use IO::Stty;

use Fcntl ':mode';
use List::Util qw[min max];

use Config::General;

use Cwd 'abs_path';
use File::Basename;

$ENV{LANG} = "C";

our $prefix;
#my $prefix = dirname(dirname(abs_path($0)));

# Begin parsing config
my $configfile = new Config::General("$prefix/etc/grid.conf");
my %config = $configfile->getall;

use Date::Parse;

if ((defined $config{GRID_PREFIX}) && $config{GRID_PREFIX} ne "") {
  $config{GRID_PREFIX} .= "/";
} else {
  $config{GRID_PREFIX} = "";
}

# --------------------------- Functions ---------------------------------------------------

sub grid_cert_install {
  # cert - certificate
  my ($cert,$key) = @_;
  system ("mkdir -p $ENV{HOME}/.globus");
  if ((defined $cert) && $cert ne "") {
    open (CERT, ">", "$ENV{HOME}/.globus/usercert.pem");
    print CERT $cert;
    close CERT;
  }
  if ((defined $key) && $key ne "") {
    open (KEY, ">", "$ENV{HOME}/.globus/userkey.pem");
    print KEY $key;
    close KEY;
  }
  system ("chmod -R 700 $ENV{HOME}/.globus");
  system ("chmod -R 600 $ENV{HOME}/.globus/*");
  return grid_cert_diag();
}


sub grid_cert_diag {
# check certificate
  my $r = `ls ~/.globus/usercert.pem 2>/dev/null`;
  if ( $r =~ m/usercert.pem/) {
    my @r = `cat ~/.globus/usercert.pem`;
    my $res = "";
    my $stop = 0;
    my %cert;
    foreach my $l (@r) {
      $stop = 1 if ($l =~ m/--BEGIN CERTIFICATE--/);
      last if $stop;
      if (my ($p) = $l =~ m/Issuer: (.+)/) {
        $cert{'issuer'} = $p;
        next;
      }
      if (my ($p) = $l =~ m/Subject: (.+)/) {
        $cert{'person'} = $p;
        next;
      }
      if (my ($p) = $l =~ m/Not Before: (.+)/) {
        $cert{'valid_from'} = str2time($p);
        next;
      }
      if (my ($p) = $l =~ m/Not After : (.+)/) {
        $cert{'valid_to'} = str2time($p);
        next;
      }
      $cert{'key'} = 1 if -s "$ENV{HOME}/.globus/userkey.pem";
    }
    return \%cert;
  } else {
    return 0;
  }
  return 0;
}


sub grid_vo_install {
  # Install virtual organization list
  my ($conf) = @_;
  system ("mkdir -p $ENV{HOME}/.voms");
  open (VO, ">", "$ENV{HOME}/.voms/voms.conf");
  print VO $conf;
  close VO;
  return grid_vo_list();
}

sub grid_vo_list {
  # List names and contence of virtual organizations files
  my @vo_l;
#  my @res = `cat $ENV{HOME}/.voms/voms.conf`;
  foreach my $l (`cat $ENV{HOME}/.voms/voms.conf`) {
#print "File: $f\n";
#    my ($n) = $f =~ m/\/([^\/]+)\.conf$/;
#    $vo_l{$n} = `cat $f`;
#    my $res = `head $f`;
    my %vo_h;
    ($vo_h{alias},$vo_h{server},$vo_h{port},$vo_h{dn},$vo_h{name}) = 
      $l =~ m/"(.+)" +"(.+)" +"(.+)" +"(.+)" +"(.+)"/;
    push @vo_l, \%vo_h if defined $vo_h{alias};
  }
  return \@vo_l;
}


sub grid_proxy_init {
  # passwd - certificate password
  my ($passwd,$vo,$my_U,$my_H) = @_;
  my $vo_str = ""; my $myproxy_str = "";
  if ((defined $vo) && ($vo ne "")) {
    $vo_str = "--vomses=$ENV{HOME}/.voms/voms.conf --voms=$vo";
  }
  if ((defined $my_U) && ($my_U ne "") && (defined $my_H) && ($my_H ne "")) {
    $myproxy_str = "-U $my_U -L $my_H -M get";
  }
  my $exp = Expect->spawn("$config{GRID_PREFIX}arcproxy $vo_str $myproxy_str")
    or die "Cannot spawn arcproxy: $!\n";

  $exp->stty(qw(raw -echo));
  $exp->log_stdout(0);
  $exp->send("$passwd\n");
  $exp->soft_close();

  return grid_proxy_info();
}

# TODO not used now
sub grid_myproxy_install {
  # passwd - certificate password
  my ($passwd,$vo,$my_U,$my_H,$my_PWD) = @_;
  my $vo_str = ""; my $myproxy_str = "";
  if ((defined $vo) && ($vo ne "")) {
    $vo_str = "--vomses=$ENV{HOME}/.voms/$vo.conf --voms=$vo";
  }
  if ((defined $my_U) && ($my_U ne "") && (defined $my_H) && ($my_H ne "")) {
    $myproxy_str = "-U $my_U -L $my_H -M put";
  }
  my $exp = Expect->spawn("$config{GRID_PREFIX}arcproxy $vo_str $myproxy_str")
    or die "Cannot spawn arcproxy: $!\n";

  $exp->stty(qw(raw -echo));
  $exp->log_stdout(0);
  $exp->send("$passwd\n");
  $exp->send("$my_PWD\n");
  $exp->soft_close();

  return "";
}


# arcproxy --voms=ukraine -U tikus -L ds7.ilt.kharkov.ua -M put
# arcproxy  -U tikus -L ds7.ilt.kharkov.ua -M get

sub grid_proxy_info {
  my $res;
  my %proxy;
  require "$prefix/libexec/modules/utils.pl";
  my @res = `$config{GRID_PREFIX}arcproxy -I 2>&1`;
  foreach my $l (@res) {
    if (my ($var,$val) = $l =~ m/^ *([^:]+) *: (.*)/) {
      if ( $var eq "Time left for proxy") {
        my $y; my $w; my $d; my $h; my $m; my $M;
        ($y) = $val =~ m/(\d+) year/;
        ($M) = $val =~ m/(\d+) month/;
        ($w) = $val =~ m/(\d+) week/;
        ($d) = $val =~ m/(\d+) day/;
        ($h) = $val =~ m/(\d+) hour/;
        ($m) = $val =~ m/(\d+) minute/;
        $y = 0 if not defined $y;
        $M = 0 if not defined $M;
        $w = 0 if not defined $w;
        $d = 0 if not defined $d;
        $h = 0 if not defined $h;
        $m = 0 if not defined $m;
        $proxy{'timeleft'} = (((($y * 365) + $M*30 + $w*7 + $d)*24 + $h)*60+$m)*60;
      } else {
        $proxy{$var} = $val;
      }
    }
  }
  return \%proxy;
}


sub grid_proxy_destroy {
  my $res;
  $res = `$config{GRID_PREFIX}arcproxy -r`;
  return grid_proxy_info();
}


# arcsub -e '&(executable="/bin/echo")(arguments="Hello World")(stdout="hello.txt")'
sub grid_launch {
  # Launch grid job
  # Parameters: 
  # H - cluster  
  # J - jobname
  # d - current dir
  # iF - input files
  # oF - output files   
  # a - app name                  
  # p - partition                 
  # n - number of CPUs            
  # t - time limit                        
  # r - runtime (rsl flag)
  # g - commandline args
  # mysql_host
  # mysql_login
  # mysql_pass
  # mysql_db
                                
  my ($H,$J,$d,$iF,$oF,$a,$e,$p,$n,$t,$r,$g,$mh,$ml,$mw,$md) = @_;
  my $cluster="";
  $cluster = "-c $H" if (defined $H && $H ne "");
  my $queue="";
  $queue = "(queue=$p)" if (defined $p && $p ne "");
  my $env="";
  $env = "(runtime=$e)" if (defined $e && $e ne "");

# Parse aruments
#  $g =~ s/ /\" \"/g;
#  $g = "\"".$g."\"";

  my $inputFiles = "";
  if (defined $iF && $iF ne ""){
    $inputFiles = "(inputFiles=";
    foreach my $i (split (",",$iF)) {
      $inputFiles .= "(\"$i\" \"\")";
    }
    $inputFiles .= ")";
  }

  my $outputFiles = "";
  if (defined $oF && $oF ne ""){
    $outputFiles = "(outputFiles=";
    foreach my $i (split (",",$oF)) {
      $outputFiles .= "(\"$i\" \"\")";
    }
    $outputFiles .= ")";
  }
	
  chdir ($d);
  my %res;
  my $res_str;
  if (defined $r && $r eq "rsl") {
# $r = `nohup $config{GRID_PREFIX}/arcsub $cluster -f \'$a\' 2>&1 /dev/null &`;
    $res_str = `$config{GRID_PREFIX}arcsub $cluster -f \'$a\'`;
  } else {
    my $arg = "";
    $arg = "(arguments=''$g'')" if ((defined $g) && ($g ne ""));
    my $l_str = "&$queue(executable=\"$a\")(jobname=\"$J\")$inputFiles$outputFiles$env$arg(stdout=\"$J.out\")(stderr=\"$J.err\")(cpuTime=\"$t\")(count=\"$n\")";
#print "\nLaunch\n$config{GRID_PREFIX}/arcsub $cluster -e \'$l_str\'\n";
    $res_str = `$config{GRID_PREFIX}arcsub $cluster -e \'$l_str\'`;
#$r = `nohup $config{GRID_PREFIX}/arcsub $cluster -e \'$l_str\' 2>&1 /dev/null &`;
#$res .= "job rsl: ". $l_str. "\n";
  }
  if (my ($p) = $res_str =~ m/(gsiftp.+)/) {
    $res{id} = $p;
    ($res{H}) = $p =~ m/\/\/(.+):/;
  } else {
    $res{error} = $res_str;
  }
  if (defined $res{id}) {
    # Connect to mysql
    my $dsn = "DBI:mysql:database=$md;host=$mh;port=3306";
    my $dbh = DBI->connect($dsn, $ml, $mw,{'RaiseError' => 1});

    my $q = "INSERT grid_tasks (task_id, hostname, task_name, work_dir, login, middleware) values (\"$res{id}\", \"$res{H}\", \"$J\", \"$d\", \"$ENV{USER}\", \"arc\");";
    $dbh->do($q);
    $dbh->disconnect();
  }
  return \%res;
}

sub grid_resources {
  my @result;
#  my @res = `cat ~/grid-resources.txt`;
  my @res = `$config{GRID_PREFIX}arcinfo -l`;
  push @res , "\n";

# TODO now using old arc
  my @allow_res = `$config{GRID_PREFIX}ngtest -R`;
#  my @allow_res = `cat ~/grid-allow.txt`;
# Parse allowed clusters
  my %allowed;
  my $cl_area = 0;
  foreach my $l (@allow_res) {
    if ($l =~ m/You are authorized at the following clusters/) {
      $cl_area = 1;
      next;
    }
    if ($cl_area && $l =~ m/^[ \t]*$/) {
      $cl_area = 0;
      next;
    }
    if ($cl_area) {
      for my $i (split(" ",$l)) {
        $allowed{$i} = 1;
      }
      next;
    }
  }

  my %resources;
  my $cluster;
  my $section = "";
  my $desc_area = 0;
		
  # Parse cluster list
  foreach my $l (@res) {
    my $p;
    my $q;
    if ( ($l =~ m/^Execution Target on Computing Service: [\w.-]+/ || 
          $l =~ m/^Execution Service: [\w.-]+/) && defined $resources{'H'}) {
# print "Commit $queue{P}, l: $l, queue_area:$queue_area\n";
# print "$resources{'Number of CPUs'}\n" if defined $resources{'Number of CPUs'};
# print "$resources{'Number of Used CPUs'}\n" if defined $resources{'Number of Used CPUs'};

        # Remove trash information
        delete $resources{'Queue'} if defined $resources{'Queue'};
        # Push all cluster information to queue
        if (defined $resources{'Installed application environments'}) {
          delete $resources{'Installed application environments'};
	}

        my $y; my $w; my $d; my $h; my $m; my $M;
        my $t_str;
					
        if (defined $resources{'Queue/ Max wall-time'}) {
          $t_str = $resources{'Queue/ Max wall-time'};
          delete $resources{'Queue/ Max wall-time'};
	}
        if (defined $resources{'Queue/ Name'}) {
          $resources{'P'} = $resources{'Queue/ Name'};
          delete $resources{'Queue/ Name'};
        }
        
        if (defined $t_str) {
          ($y) = $t_str =~ m/(\d+) year/;
          ($M) = $t_str =~ m/(\d+) month/;
          ($w) = $t_str =~ m/(\d+) week/;
          ($d) = $t_str =~ m/(\d+) day/;
          ($h) = $t_str =~ m/(\d+) hour/;
          ($m) = $t_str =~ m/(\d+) minute/;
        }
        $y = 0 if not defined $y;
        $M = 0 if not defined $M;
        $w = 0 if not defined $w;
        $d = 0 if not defined $d;
        $h = 0 if not defined $h;
        $m = 0 if not defined $m;
        $resources{'l'} = (((($y * 365) + $M*30 + $w*7 + $d)*24 + $h)*60+$m)*60;
        if ( ((defined $resources{'Endpoint/ Health state info'}) && $resources{'Endpoint/ Health state info'} eq 'active') ||
             ((defined $resources{'Endpoint/ Health State Info'}) && $resources{'Endpoint/ Health State Info'} eq 'active')) {
          if ($allowed{$resources{'H'}}) {
            $resources{'a'} = 'up';
          } else {
            $resources{'a'} = 'down';
          }
        } else {
          $resources{'a'} = 'fail';
        }
#        delete $queue{'Health state info'};
					
        # Calculate Nodes
#        $resources{'D'} = 0;
#        if (defined $resources{'Number of Computers'}) {
#          for my $i (split(/\n/, $resources{'Number of Computers'})) {
#            my ($n) = $i =~ m/(\d+)$/;
#            $resources{'D'} += $n;
#          }
#        }

        # Make idle queue
        $resources{'t'} = 'idle';
#        $resources{'C'} = max($resources{'Number of CPUs'} - $resources{'Number of Used CPUs'}, 0);
        if (defined  $resources{'Manager/ Total logical CPUs'}) {
          $resources{'C'} = $resources{'Manager/ Total logical CPUs'};
          delete $resources{'Manager/ Total logical CPUs'};
        }
	# TODO This is incorrect number
        if (defined  $resources{'Manager/ Total physical CPUs'}) {
          $resources{'D'} = $resources{'Manager/ Total physical CPUs'};
          delete $resources{'Manager/ Total physical CPUs'};
        }
        my %q_i = %resources;
#        # Make alloc queue
#        $resources{'t'} = 'alloc';
#        $resources{'C'} = $resources{'Number of Used CPUs'};
#        my %q_a = %resources;
        push @result, \%q_i;
#        push @result, \%q_a;
#        %resources = ();
    } # End of pushing cluster information
    
    # Process empty line
    if ($l =~ m/^$/) {
      $desc_area = 0;
      $section = "";
      next;
    }    

    # Process new cluster 
    if (($p) = $l =~ m/^Execution Target on Computing Service: ([\w.-]+)/ ){
      $cluster = $p;
      %resources = ();
      $resources{'H'} = $p;
      $desc_area = 0;
      next;
    }
    # Process new cluster Variant 2  
    if (($p) = $l =~ m/^Execution Service: ([\w.-]+)/ ){
      $cluster = $p;
      %resources = ();
      $resources{'H'} = $p;
      $desc_area = 0;
      next;
    }


    # Process section line
    if (($p) = $l =~ m/^([\w.]+) information\:/ ){
      $section = $p."/ ";
      $desc_area = 0;
      next;
    }

    next if not defined $cluster;

    # Adding cluster properties
    if ( ($p,$q) = $l =~ m/^ +(.+): +(.+)/) {
      $resources{$section.$p} = $q;
      $desc_area = 0;
      next;
    }
    if (($p) = $l =~ m/^ +(.+):$/) {
#print "Start multiline $p\n";
      $desc_area = $p;
      next;
    }
    if ($desc_area && (($p) = $l =~ m/^ +(.+)/)) {
#print "Continue multiline $desc_area with $p\n";
      $resources{$desc_area} = "" if not defined $resources{$desc_area};
      $resources{$desc_area} .= $p."\n";
      next;
    }
    # Adding cluster flags
    if (!$desc_area && (($p) = $l =~ m/^ +([\w ']+)/)) {
      $resources{$section.$p} = 1;
      $desc_area = 0;
      next;
    }
  } # End of foreach
  return \@result;
}


sub grid_queue {
  # parameter detail - detailed view
  my ($detail) = @_;
  my $params;
  if (defined $detail) {
    $params = "-a -l";
  } else {
    $params = "-l -s PREPARED -s ACCEPTED -s \"INLRMS: Q\" -s \"INLRMS: R\" -s SUBMITTING -s CANCELING -s FINISHING";
  }
  my @result;
  my @res = `$config{GRID_PREFIX}arcstat $params`;
  my %job;
  foreach my $l (@res) {
    if ( my ($p) = $l =~ m/^Job: (.+)/) {
      if (defined $job{'i'}) {
        # Push job to results
        my %j_t = %job;
        push @result, \%j_t;
      }
      %job = ();
      $job{'i'} = $p;
#print "Found job $job{'id'}\n";
      next;
    }
    if (defined $job{'i'} && (my ($p,$q) = $l =~ m/ +([^:]+): (.*)/)) {
#print "P: $p, $q\n";
      if ($p eq 'Name') {
        $job{'j'} = $q;
      } elsif ($p eq 'State') {
        $job{'t'} = 'CD'  if ($q =~ m/FINISHED/);
        $job{'t'} = 'F'   if ($q =~ m/FAILED/);
        $job{'t'} = 'PD'  if (($q =~ m/INLRMS: *Q/) || ($q =~ m/PREPARED/) || ($q =~ m/ACCEPTED/));
        $job{'t'} = 'R'   if ($q =~ m/INLRMS: *R/);
        $job{'t'} = 'CG'  if (($q =~ m/CANCELING/) || ($q =~ m/FINISHING/));
        $job{'t'} = 'CA'  if ($q =~ m/KILLED/);
        $job{'t'} = 'TO'  if ($q =~ m/DELETED/);
      } elsif ($p eq 'Job Error') { 
        $job{'r'} = $q;
      } elsif ($p eq 'Cluster') { 
        ($job{'H'}) = $q =~m/\/\/([\w.-]+):/;
      } elsif ($p eq 'Owner') { 
        ($job{'u'}) = $q =~m/CN=(.+)/;
      } elsif ($p eq 'Queue') { 
        $job{'P'} = $q;
      } elsif ($p eq 'Requested Slots') { 
        $job{'C'} = $q;
      } elsif ($p eq 'Submitted') { 
        $job{'b'} = str2time($q);
      } elsif ($p eq 'End Time') { 
        $job{'e'} = str2time($q);
      } else {
        $job{$p} = $q;
      }
      next;	
    }
  }
  if (defined $job{'i'}) {
    # Push last job to results
    my %j_t = %job;
    push @result, \%j_t;
  }
  return \@result;
}


sub grid_ls {
  my ($dir) = @_;
  my @list;
  my @res = `$config{GRID_PREFIX}arcls -l $dir`;
  foreach my $l (@res) {
    my %f; my $d; my $date;
    if (($f{'f'},$d,$f{'s'},$date) = $l =~ m/([^ ]+) ([^ ]+) (\d+) ([\d\-]+ [\d\:]+)/ ) {
      $f{'s'} = int $f{'s'};
      $f{'t'} = str2time($date);
      $f{'d'} = 1;
      $f{'d'} = 0 if ( $d eq "file");
      push @list, \%f if ($f{'f'} ne "");
    }
  }
  return \@list;
}

# TODO
#  in arc 1.2 there is bug in arccp which don't work with hydra hctl

sub grid_less {
  my($f,$o,$s) = @_;
  use File::Temp qw/ :POSIX /;
  my $tf = tmpnam();
  my $res = `$config{GRID_PREFIX}ngcp $f $tf 2>&1`;
  my $r = less($tf,$o,$s);

  system("rm -f $tf");
  return $r;
}


sub grid_put {
  my($f) = @_;
  use File::Temp qw/ :POSIX /;

  my $tf = tmpnam();
  open (FILE, ">", $tf);
  while (<STDIN>) {
    print FILE $_;
  }
  my $res = `$config{GRID_PREFIX}ngcp $tf $f 2>&1`;
  system ("rm -f $tf");
  print $res;
}


sub grid_get {
  my($d,$jid) = @_;
  my $res = `$config{GRID_PREFIX}arcget --dir=$d $jid 2>&1`;
  return $res;
}

sub grid_cancel {
  my($jid) = @_;
  my $res = `$config{GRID_PREFIX}arckill -k $jid`;
}


sub grid_cp {
  my($sf_ref,$ddir) = @_;
  my $res = "";
  foreach my $f (@$sf_ref) {
    my ($fn) = $f =~ m/([^\/]+)$/;
    $res .= `$config{GRID_PREFIX}ngcp $f $ddir/$fn 2>&1`
  }
  return $res;
}

# For grid_jobget
sub grid_get_str {
  my($d,$jid) = @_;
  return "$config{GRID_PREFIX}arcget --dir=$d $jid 2>&1";
}

sub grid_clean_str {
  my($jid) = @_;
  return "$config{GRID_PREFIX}arcclean -f $jid";
}

1;
